"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ImageCdnUrlSyntax = exports.ImageCdnImplementation = void 0;
exports.prepareImageCdn = prepareImageCdn;
/* eslint-disable max-lines */
const crypto_1 = __importDefault(require("crypto"));
const module_1 = require("module");
const path_1 = require("path");
const fs_extra_1 = require("fs-extra");
const cache_1 = require("./cache");
const config_1 = require("./config");
// eslint-disable-next-line no-shadow
var ImageCdnImplementation;
(function (ImageCdnImplementation) {
    ImageCdnImplementation["NETLIFY_IMAGE_CDN"] = "NETLIFY_IMAGE_CDN";
    ImageCdnImplementation["IPX"] = "IPX";
    ImageCdnImplementation["NONE"] = "NONE";
})(ImageCdnImplementation || (exports.ImageCdnImplementation = ImageCdnImplementation = {}));
function selectImageCdnImplementation({ netlifyConfig, }) {
    /* eslint-disable no-param-reassign */
    const { NETLIFY_IMAGE_CDN, GATSBY_CLOUD_IMAGE_CDN } = netlifyConfig.build.environment;
    const imageCDNEnabled = GATSBY_CLOUD_IMAGE_CDN === '1' || GATSBY_CLOUD_IMAGE_CDN === 'true';
    if (NETLIFY_IMAGE_CDN === 'true') {
        if (!imageCDNEnabled) {
            // we do need to ensure GATSBY_CLOUD_IMAGE_CDN env var for Gatsby to not
            // transform images at build time
            netlifyConfig.build.environment.GATSBY_CLOUD_IMAGE_CDN = 'true';
        }
        return ImageCdnImplementation.NETLIFY_IMAGE_CDN;
    }
    if (imageCDNEnabled) {
        return ImageCdnImplementation.IPX;
    }
    return ImageCdnImplementation.NONE;
    /* eslint-enable no-param-reassign */
}
exports.ImageCdnUrlSyntax = {
    QUERY_WITH_ENCRYPTED_URL: 'QUERY_WITH_ENCRYPTED_URL',
    QUERY: 'QUERY',
    BASE64: 'BASE64',
    NONE: 'NONE',
};
function getGatsbyPluginUtilsSyntax(contextRequire) {
    try {
        const urlGeneratorExports = contextRequire('gatsby-plugin-utils/polyfill-remote-file/utils/url-generator');
        if (urlGeneratorExports.ImageCDNUrlKeys) {
            return urlGeneratorExports.ImageCDNUrlKeys.ENCRYPTED_URL
                ? exports.ImageCdnUrlSyntax.QUERY_WITH_ENCRYPTED_URL
                : exports.ImageCdnUrlSyntax.QUERY;
        }
        return exports.ImageCdnUrlSyntax.BASE64;
    }
    catch { }
    return exports.ImageCdnUrlSyntax.NONE;
}
// list of known packages using Gatsby Image CDN that don't use their dedicated image transformation service
// compiled through manual inspection of https://github.com/search?q=addRemoteFilePolyfillInterface&type=code
const gatsbySourcePluginsUsingImageCdn = [
    // gatsby owned
    'gatsby-source-drupal',
    'gatsby-source-wordpress',
    // 3rd party
    'gatsby-source-sanity',
    'gatsby-plugin-remote-images',
    'gatsby-source-youtube-oembed',
    'gatsby-source-airtable-next',
    'gatsby-source-directus',
];
function detectImageCdnUrlPatterns({ publish, imageCDNImplementation, }) {
    const detectedSyntaxes = new Set();
    const siteRequire = (0, module_1.createRequire)(`${(0, config_1.getGatsbyRoot)(publish)}/:internal:`);
    try {
        detectedSyntaxes.add(getGatsbyPluginUtilsSyntax(siteRequire));
    }
    catch { }
    for (const packageUsingGatsbyImageCdn of gatsbySourcePluginsUsingImageCdn) {
        try {
            const packageRequire = (0, module_1.createRequire)(siteRequire.resolve(packageUsingGatsbyImageCdn));
            detectedSyntaxes.add(getGatsbyPluginUtilsSyntax(packageRequire));
        }
        catch { }
    }
    // we only have NONE, because we need to return something from detection helper, so let's remove it from the list
    detectedSyntaxes.delete(exports.ImageCdnUrlSyntax.NONE);
    if (detectedSyntaxes.size === 0) {
        // automatic detection is not ideal, so it might not find gatsby-plugin-utils even it it is actually used
        // in this case we fall back to the most recent and safe one
        detectedSyntaxes.add(exports.ImageCdnUrlSyntax.QUERY_WITH_ENCRYPTED_URL);
    }
    if (imageCDNImplementation === ImageCdnImplementation.NETLIFY_IMAGE_CDN &&
        detectedSyntaxes.has(exports.ImageCdnUrlSyntax.QUERY_WITH_ENCRYPTED_URL)) {
        // Netlify Image CDN can't handle encrypted URLs, but it uses allowlist instead, so we replace QUERY_WITH_ENCRYPTED_URL with QUERY
        // Remote urls Allow list is required form Netlify Image CDN to work, so this should ensure no unauthorized remote urls can be
        // used successfully
        detectedSyntaxes.delete(exports.ImageCdnUrlSyntax.QUERY_WITH_ENCRYPTED_URL);
        detectedSyntaxes.add(exports.ImageCdnUrlSyntax.QUERY);
    }
    return [...detectedSyntaxes];
}
const IMAGE_CDN_ENCRYPTION_SECRET_KEY_LENGTH = 32;
const IMAGE_CDN_ENCRYPTION_IV_LENGTH = 16;
function setupImageCdnUrlEncryption({ previousImageCdnConfig, netlifyConfig, imageCDNImplementation, enabledUrlPatterns, }) {
    const shouldUseUrlEncryption = imageCDNImplementation === ImageCdnImplementation.IPX &&
        enabledUrlPatterns.includes(exports.ImageCdnUrlSyntax.QUERY_WITH_ENCRYPTED_URL);
    if (!shouldUseUrlEncryption) {
        return;
    }
    // if environment variables are already set - use them
    if (netlifyConfig.build.environment.IMAGE_CDN_ENCRYPTION_SECRET_KEY &&
        netlifyConfig.build.environment.IMAGE_CDN_ENCRYPTION_IV) {
        return {
            IMAGE_CDN_ENCRYPTION_SECRET_KEY: netlifyConfig.build.environment.IMAGE_CDN_ENCRYPTION_SECRET_KEY,
            IMAGE_CDN_ENCRYPTION_IV: netlifyConfig.build.environment.IMAGE_CDN_ENCRYPTION_IV,
        };
    }
    let encryptionConfig = previousImageCdnConfig === null || previousImageCdnConfig === void 0 ? void 0 : previousImageCdnConfig.encryptionConfig;
    if (!encryptionConfig) {
        // if we don't yet have encryption config, generate new one
        encryptionConfig = {
            IMAGE_CDN_ENCRYPTION_SECRET_KEY: crypto_1.default
                .randomBytes(IMAGE_CDN_ENCRYPTION_SECRET_KEY_LENGTH)
                .toString(`hex`),
            IMAGE_CDN_ENCRYPTION_IV: crypto_1.default
                .randomBytes(IMAGE_CDN_ENCRYPTION_IV_LENGTH)
                .toString(`hex`),
        };
    }
    // eslint-disable-next-line no-param-reassign
    netlifyConfig.build.environment.IMAGE_CDN_ENCRYPTION_SECRET_KEY =
        encryptionConfig.IMAGE_CDN_ENCRYPTION_SECRET_KEY;
    // eslint-disable-next-line no-param-reassign
    netlifyConfig.build.environment.IMAGE_CDN_ENCRYPTION_IV =
        encryptionConfig.IMAGE_CDN_ENCRYPTION_IV;
    return encryptionConfig;
}
// eslint-disable-next-line max-statements
async function prepareImageCdn({ publish, netlifyConfig, }) {
    var _a, _b;
    const cacheDir = (0, cache_1.normalizedCacheDir)(publish);
    // private location that is cached between builds, but not exposed to public
    const imageCdnConfigFile = (0, path_1.join)(cacheDir, 'imageCdnEncryptionConfig.json');
    let previousImageCdnConfig;
    try {
        previousImageCdnConfig = await (0, fs_extra_1.readJSON)(imageCdnConfigFile);
    }
    catch { }
    const imageCDNImplementation = await selectImageCdnImplementation({
        netlifyConfig,
    });
    const enabledUrlPatterns = await detectImageCdnUrlPatterns({
        publish,
        imageCDNImplementation,
    });
    const encryptionConfig = await setupImageCdnUrlEncryption({
        previousImageCdnConfig,
        netlifyConfig,
        imageCDNImplementation,
        enabledUrlPatterns,
    });
    const allowedRemoteUrls = 
    // use defined remote images if user configured them
    ((_b = (_a = netlifyConfig.images) === null || _a === void 0 ? void 0 : _a.remote_images) !== null && _b !== void 0 ? _b : 
    // if they are not configured and user
    //  - is using Netlify Image CDN - we should use empty array to disallow any remote url
    //  - is using IPX - we should not setup remote urls as IPX/Gatsby Cloud never required them and this would be breaking change
    imageCDNImplementation === ImageCdnImplementation.NETLIFY_IMAGE_CDN)
        ? []
        : undefined;
    const imageCDNConfig = {
        enabledUrlPatterns,
        encryptionConfig,
        allowedRemoteUrls,
    };
    // some configuration change require to throw away previous build artifacts to prevent gatsby from reusing
    // previous html/json files that would contain old urls that would not work with new configuration
    // so we need to ensure fresh builds when at least one of the following changed since last build:
    // - enabledUrlPatterns changed
    // - encryption changed
    const enabledUrlPatternsChanged = JSON.stringify(previousImageCdnConfig === null || previousImageCdnConfig === void 0 ? void 0 : previousImageCdnConfig.enabledUrlPatterns) !==
        JSON.stringify(imageCDNConfig.enabledUrlPatterns);
    const encryptionConfigChanged = JSON.stringify(previousImageCdnConfig === null || previousImageCdnConfig === void 0 ? void 0 : previousImageCdnConfig.encryptionConfig) !==
        JSON.stringify(imageCDNConfig.encryptionConfig);
    if (enabledUrlPatternsChanged || encryptionConfigChanged) {
        const dirsToClean = (0, cache_1.getCacheDirs)(publish).filter(fs_extra_1.existsSync);
        if (dirsToClean.length !== 0) {
            console.log(`Image CDN configuration changed:${enabledUrlPatternsChanged ? `\n - enabledUrlPatterns` : ''}${encryptionConfigChanged ? `\n - encryptionConfig` : ''}\nCleaning previous build and building fresh.`);
            for (const dirToClean of dirsToClean) {
                await (0, fs_extra_1.remove)(dirToClean);
            }
        }
    }
    return {
        imageCDNConfig,
        imageCDNImplementation,
        postBuildCallback: async () => {
            // .cache directory might not exist yet, so to not mess with Gatsby itself
            // we delay storing the config for future build until after the build
            await (0, fs_extra_1.writeJSON)(imageCdnConfigFile, imageCDNConfig);
        },
    };
}
/* eslint-enable max-lines */
