"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var _a;
Object.defineProperty(exports, "__esModule", { value: true });
exports.handler = void 0;
const buffer_1 = require("buffer");
const crypto_1 = __importDefault(require("crypto"));
const ipx_1 = require("@netlify/ipx");
/**
 * @type {import('./config').IpxWrapperConfig}
 */
const config_1 = require("./config");
const rawConfig = config_1.config;
const config = {
    ...rawConfig,
    remoteUrls: (_a = rawConfig.allowedRemoteUrls) === null || _a === void 0 ? void 0 : _a.map((remoteUrl) => new RegExp(remoteUrl)),
};
const ipxHandler = (0, ipx_1.createIPXHandler)({
    propsEncoding: 'base64',
    basePath: '/_gatsby/image/',
    bypassDomainCheck: true,
});
function decryptUrl(url) {
    if (config.encryptionConfig) {
        const key = buffer_1.Buffer.from(config.encryptionConfig.IMAGE_CDN_ENCRYPTION_SECRET_KEY, `hex`);
        const iv = buffer_1.Buffer.from(config.encryptionConfig.IMAGE_CDN_ENCRYPTION_IV, `hex`);
        const decipher = crypto_1.default.createDecipheriv(`aes-256-ctr`, key, iv);
        const decrypted = decipher.update(buffer_1.Buffer.from(url, `hex`));
        const clearText = buffer_1.Buffer.concat([decrypted, decipher.final()]).toString();
        const [
        // padding is delimited by a ":"
        // eslint-disable-next-line @typescript-eslint/no-unused-vars
        _randomPadding, 
        // urls also contain ":"
        ...urlParts] = clearText.split(`:`);
        const decodedUrl = urlParts.join(`:`);
        return decodedUrl;
    }
    throw new Error(`No encryption config found`);
}
const QUERY_PARAM_IMAGE_REDIRECT_URL = /^\/\.netlify\/builders\/_ipx\/image_query_compat(?<eu>_eu)?\/(?<args>[^/]+?)\/(?<url>[^/]+?)\/([^/]+?)\/?$/i;
const QUERY_PARAM_FILE_REDIRECT_URL = /^\/\.netlify\/functions\/_ipx\/file_query_compat(?<eu>_eu)?\/(?<url>[^/]+?)\/([^/]+?)\/?$/i;
function matchRequestTypeAndArguments(event) {
    const { pathname, host } = new URL(event.rawUrl);
    const queryParamImageMatch = pathname.match(QUERY_PARAM_IMAGE_REDIRECT_URL);
    if (queryParamImageMatch === null || queryParamImageMatch === void 0 ? void 0 : queryParamImageMatch.groups) {
        const url = queryParamImageMatch.groups.eu
            ? decryptUrl(queryParamImageMatch.groups.url)
            : decodeURIComponent(queryParamImageMatch.groups.url);
        const base64EncodedUrl = buffer_1.Buffer.from(url, 'utf8').toString('base64');
        const base64EncodedArgs = buffer_1.Buffer.from(decodeURIComponent(queryParamImageMatch.groups.args), 'utf8').toString('base64');
        return {
            type: `image`,
            url,
            event: {
                ...event,
                path: `/_gatsby/image/${base64EncodedUrl}/${base64EncodedArgs}`,
            },
            urlPattern: queryParamImageMatch.groups.eu
                ? 'QUERY_WITH_ENCRYPTED_URL'
                : 'QUERY',
        };
    }
    const queryParamFileMatch = pathname.match(QUERY_PARAM_FILE_REDIRECT_URL);
    if (queryParamFileMatch === null || queryParamFileMatch === void 0 ? void 0 : queryParamFileMatch.groups) {
        const url = queryParamFileMatch.groups.eu
            ? decryptUrl(queryParamFileMatch.groups.url)
            : decodeURIComponent(queryParamFileMatch.groups.url);
        return {
            type: `file`,
            url,
            originalHost: host,
            urlPattern: queryParamFileMatch.groups.eu
                ? 'QUERY_WITH_ENCRYPTED_URL'
                : 'QUERY',
        };
    }
    const [, , type, encodedUrl] = pathname.split('/');
    const url = buffer_1.Buffer.from(encodedUrl, 'base64').toString('utf8');
    if (type === `file`) {
        return {
            type: `file`,
            url,
            originalHost: host,
            urlPattern: 'BASE64',
        };
    }
    if (type === `image`) {
        return {
            type: `image`,
            url,
            event,
            urlPattern: 'BASE64',
        };
    }
    // no known matching
    return undefined;
}
// eslint-disable-next-line require-await
const handler = async (event, ...rest) => {
    const requestTypeAndArgs = matchRequestTypeAndArguments(event);
    if (!requestTypeAndArgs) {
        return {
            statusCode: 400,
            body: 'Invalid request',
        };
    }
    if (!config.enabledUrlPatterns.includes(requestTypeAndArgs.urlPattern)) {
        return {
            statusCode: 400,
            body: 'Invalid request',
        };
    }
    if (config.remoteUrls &&
        !config.remoteUrls.some((remoteUrl) => remoteUrl.test(requestTypeAndArgs.url))) {
        return {
            statusCode: 400,
            body: 'Invalid request',
        };
    }
    if (requestTypeAndArgs.type === `image`) {
        return ipxHandler(requestTypeAndArgs.event, ...rest);
    }
    try {
        const urlString = requestTypeAndArgs.url;
        // Validate it by parsing it
        const url = new URL(urlString);
        if (url.host === requestTypeAndArgs.originalHost) {
            return {
                statusCode: 400,
                body: 'File cannot be served from the same host as the original request',
            };
        }
        console.log(`Redirecting to ${urlString}`);
        return {
            statusCode: 301,
            headers: {
                Location: url.toString(),
            },
            body: '',
        };
    }
    catch (error) {
        return {
            statusCode: 400,
            body: JSON.stringify({
                msg: `Redirecting to ${requestTypeAndArgs.url} failed`,
                error,
            }),
        };
    }
};
exports.handler = handler;
